/*!
	\file NURBSSurface.h
	\brief TrimmedNURBSSurface class declaration

	Information :
		- Author	: Vincent PRAT (vprat@ifrance.com)
		- Created	: 10-07-2001 (dd-mm-yyyy)
		- Modified	: 13-07-2001 (dd-mm-yyyy)

	Copyright  2001 Vincent PRAT
*/

#ifndef __TRIMMED_NURBS_SURFACE_H__
#define __TRIMMED_NURBS_SURFACE_H__

#ifdef WIN32
#include <windows.h>	// Used by gl.h
#endif

#ifndef WIN32
#include <string.h>		// memset
#endif

#include <fstream.h>	// ifstream
#include <GL/gl.h>		// GL Stuff (glVertex, glBegin, ...)
#include <math.h>		// sin, cos
#include "VectorNd.h"	// Vector2d, Vector3d and Vector4d
#include "Contour2D.h"	// Contour2D

/*!
	\brief Default number of vertices generated for one surface, in each direction
*/
#define DEFAULT_TESSELATION_VALUE 20

/*!
	\brief Distance after which the surface is tesselated with the minimal tesselation value

	Used for Dynamic tesselation
*/
#define FAR_DISTANCE 60

/*!
	\brief Distance before which the surface is tesselated with the default tesselation value

	Used for Dynamic tesselation
*/
#define CLOSE_DISTANCE 10

/*!
	\brief Object representing a trimmed NURBS Surface

	This surface is trimmed by NURBS curves (Contour2D objects) defined
	in parameter plane (u, v). We read the surface from a file.

	Some cool things are missing :
		- Modify control points to enable surface deformation
		- Dynamic tesselation based on the distance to user AND the volume of
		the surface.
*/
class TrimmedNURBSSurface
{
	// Not really a nice thing but so handy for the function
	// Contour2D::ComputeSurfacePoints( TrimmedNURBSSurface *surface ) !
	friend	class Contour2D;

protected:	
	/*!
		\brief current position of the surface in world space.
	*/
	Vector3d vPosition;

	/*!
		\brief order of the surface in U and V direction

		Element 0 of this array is for U direction
		Element 1 of this array is for V direction

		These integers must be strictly higher than 1 (>=2)
	*/
	int		anOrder[2];

	/*!
		\brief Number of control points defining the surface

		Element 0 of this array is for U direction
		Element 1 of this array is for V direction
	*/
	int		anNumberOfControlPoints[2];	
	
	/*!
		\brief degree of the surface in U and V direction

		Element 0 of this array is for U direction
		Element 1 of this array is for V direction
	*/
	int		anDegree[2];				
	
	/*!
		\brief Number of values in the knot span

		This is deduced from other constants :
		numberOfKnots = numberOfControlPoints + order

		Element 0 of this array is for U direction
		Element 1 of this array is for V direction
	*/
	int		anNumberOfKnotSpans[2];		
	
	/*!
		\brief order of the surface in U and V direction

		Element 0 of this array is for U direction
		Element 1 of this array is for V direction
	*/
	int		anNumberOfKnots[2];
	
	/*!
		\brief Current tesselation value of the curve

		This represents the number of segments that will approximate the
		mathematical curve. Thus, there are tesselation+1 vertices.
	*/
	int		nTesselation;				

	/*!
		\brief Number of NURBS curves trimming the surface
	*/
	int		nNumberOfContours;
	
	/*!
		\brief Indicates whether we shall trim the surface

		This can be used to cancel trimming when surface is too far for instance, 
		thus saving costly computations
	*/
	bool	bPerformTrimming;
	
	/*!
		\brief pointer to U knot vector

		The size of the array is numberOfKnots[0]
	*/
	float*	pfUKnots;
	
	/*!
		\brief pointer to coefficients of the basis function polynomials corresponding to U direction

		These values are pre-calculated at surface creation only. 
		The size of the array is order[0]*order[0]*numberOfKnotSpans[0]
	*/
	float*	pfUBasisCoefficients;					
	
	/*!
		\brief pointer to values of the basis function corresponding to U direction

		These values are pre-calculated each time the tesselation value changes.
		The size of the array is tesselation+1
	*/
	float*	pfUBasisValues;	
	
	/*!
		\brief pointer to values of the derivative of the basis function corresponding to U direction

		These values are pre-calculated each time the tesselation value changes.
		The size of the array is tesselation+1
	*/
	float*	pfDUBasisValues;	

	/*!
		\brief Pointer to the knot span corresponding to U direction

		The size of the array is tesselation+1
	*/
	int*	pnUTessKnotSpan;	
	
	/*!
		\brief pointer to V knot vector

		The size of the array is numberOfKnots[1]
	*/
	float*	pfVKnots;		

	/*!
		\brief pointer to coefficients of the basis function polynomials corresponding to V direction

		These values are pre-calculated at surface creation only. 
		The size of the array is order[1]*order[1]*numberOfKnotSpans[1]
	*/
	float*	pfVBasisCoefficients;	
	
	/*!
		\brief pointer to values of the basis function corresponding to V direction

		These values are pre-calculated each time the tesselation value changes.
		The size of the array is tesselation+1
	*/
	float*	pfVBasisValues;	

	/*!
		\brief pointer to values of the derivative of the basis function corresponding to V direction

		These values are pre-calculated each time the tesselation value changes.
		The size of the array is tesselation+1
	*/
	float*	pfDVBasisValues;

	/*!
		\brief Pointer to the knot span corresponding to V direction

		The size of the array is tesselation+1
	*/
	int*	pnVTessKnotSpan;
	
	/*!
		\brief Pointer to the array of control points

		Control points are stored linearly, firstly incrementing U value and then V value.
		The size of the array is numberOfControlPoints[0]*numberOfControlPoints[1]
	*/
	Vector4d*	pControlPoints;	

	/*!
		\brief Pointer to tesselated surface points (vertex buffer)

		Those points are in the object space (not transformed)
		The size of the array is (tesselation+1)*(tesselation+1)
	*/
	Vector4d*	pVertexBuffer;		

	/*!
		\brief Pointer to tesselated surface points normals

		The size of the array is (tesselation+1)*(tesselation+1)
	*/
	Vector3d*	pNormals;		

	/*!
		\brief Pointer to tesselated surface points texture coordinates

		We use a uniform planar texture mapping method.
		The size of the array is (tesselation+1)*(tesselation+1)
	*/
	Vector2d*	pTextureCoordinates;
	
	/*!
		\brief Pointer to NURBS Curves trimming the surface

		The size of the array is nNumberOfContours
	*/
	Contour2D**	pContours;

	/*!
		\brief Initialises and computes every variable that is not provided during construction.

		This calculates for instance the degree, the number of knots and such deduced 
		values. Also pre-calculates basis coefficients, basis values and launches the 
		first tesselation.
	*/
	void Initialise();

	/*!
		\brief Computes one particular polynomial coefficient (which depends only on the knotspan we are in)
		\param  nInterval knot vector interval for this particular coefficient
		\param i see formula
		\param p see formula
		\param k see formula

		This function is recursive and uses the formula :
		\image html formula01.gif
	*/
	float ComputeCoefficient( float *pfKnots, int nInterval, int i, int p, int k );

	/*!
		\brief Fills the pfUBasisCoefficients and pfVBasisCoefficients arrays.

		This function uses the ComputeCoefficient function to find the particular value
		of a certain coefficient. Function called only once by Initialise().
	*/
	void EvaluateBasisFunctionsCoefficients();

	/*!
		\brief Fills the pfUBasisValues and pfVBasisValues arrays

		This function calculates the basis values for each tesselated point. At the
		same time, we also fill the pnTessKnotSpan array. We use the formula :
		\image html formula02.gif
	*/
	void EvaluateBasisFunctions();

	/*!
		\brief Evaluates the basis function for a particular couple of parameter values

		We use the recursive formula \image html formula03.gif
	*/
	float EvaluateBasisFunction( int i, int p, float u, float *pfKnots );

	/*!
		\brief Tesselates the surface

		We calculate the surface points, normals and texture coordinates. Then, we trim the surface.
	*/
	void Tesselate();

public:
	/*!
		\brief Constructor of a TrimmedNURBSSurface object
		\param pcFileName Pointer to the name of the file to read from

		The paramters of the surface are read in a file. The file format is the following : <BR>
			Order					[int int]<BR>
			NumberOfControlPoints 	[int int]<BR>
			ControlPoint0 			[float float float float] (ie [x y z w])<BR>
			ControlPoint1 			[float float float float]<BR>
			...<BR>
			UKnot0					[float]<BR>
			UKnot1					[float]<BR>
			...<BR>
			VKnot0					[float]<BR>
			VKnot1					[float]<BR>
			...<BR>
			Number of contours		[int]<BR>
			Contour0 				[see contour format]<BR>
			Contour1				[see contour format]<BR>
			...<BR>
<BR>
		The format for a contour is :<BR>
			Order					[int]<BR>
			NumberOfControlPoints	[int]<BR>
			ControlPoint0			[float float float] (ie [u v w])<BR>
			ControlPoint1			[float float float] (ie [u v w])<BR>
			...<BR>
			Knot0					[float]<BR>
			Knot1					[float]<BR>
			...<BR>
	*/
	TrimmedNURBSSurface( char *pcFileName );

	/*!
		\brief Destructor of the object

		Frees all the memory that was allocated dynamically
	*/
	virtual ~TrimmedNURBSSurface();

	/*!
		\brief Change current tesselation value
	*/
	void SetTesselation( int newValue );

	/*!
		\brief Render the NURBS surface using dynamic tesselation
		\param xCam current x coordinate of the camera (used to calculate distance to object)
		\param xCam current y coordinate of the camera (used to calculate distance to object)
		\param xCam current z coordinate of the camera (used to calculate distance to object)

		We build surface mesh from the vertex buffer filled by Tesselate(). We use triangle
		strips to lower used memory.
		Dynamic tesselation is based only on distance to viewer. We should also take into account
		the volume of the surface.
		This function uses OpenGL, create a subclass inheriting of this one and redefining this
		function to use another 3D API.
	*/
	virtual void RenderLOD( float xCam, float yCam, float zCam );

	/*!
		\brief Render the NURBS surface without using dynamic tesselation

		We build surface mesh from the vertex buffer filled by Tesselate(). We use triangle
		strips to lower used memory.
		This function uses OpenGL, create a subclass inheriting of this one and redefining this
		function to use another 3D API.
	*/
	virtual void Render();

	/*!
		\brief Saves current matrix
		
		This function should be called once before any call to Translate, Scale or Rotate.
	*/
	virtual void Begin() 
	{
		// We assume that the surface is originally centered on the origin
		memset( &vPosition, 0, sizeof( Vector3d ) );

		// Save current matrix
		glPushMatrix(); 
	}

	/*!
		\brief Restore saved matrix

		This function should be called if Begin() was called.
	*/
	virtual void End() 
	{ 
		// Restore saved matrix
		glPopMatrix(); 
	}

	/*!
		\brief Translates the surface
		\param x distance to translate on X axis
		\param y distance to translate on Y axis
		\param z distance to translate on Z axis
		
		This function uses OpenGL.
    */
	virtual void Translate( float x, float y, float z ) 
	{ 
		// OpenGL translation
		glTranslatef( x, y, z );

		// Update surface position
		vPosition.x += x;
		vPosition.y += y;
		vPosition.z += z;
	}
	
	/*!
		\brief Rotates the surface
		\param x angle to rotate around X axis
		\param y angle to rotate around Y axis
		\param z angle to rotate around Z axis
		
		This function uses OpenGL.
    */
	virtual void Rotate( float x, float y, float z ) 
	{
		// OpenGL rotation
		glRotatef( x, 1, 0, 0 );
		glRotatef( y, 0, 1, 0 );
		glRotatef( z, 0, 0, 1 );

		// We should update surface position here but I did not have the 
		// time to find the right method. We will not have a correct position
		// If we apply any translation prior to a rotation.
	}

	/*!
		\brief Scales the surface
		\param x scale factor on X axis
		\param y scale factor on Y axis
		\param z scale factor on Z axis
		
		This function uses OpenGL.
    */
	virtual void Scale( float x, float y, float z )
	{
		glScalef( x, y, z );
	}
};

#endif