/*!
	\file NURBSSurface.cpp
	\brief TrimmedNURBSSurface class methods definition and implementation

	Information :
		- Author	: Vincent PRAT (vprat@ifrance.com)
		- Created	: 10-07-2001 (dd-mm-yyyy)
		- Modified	: 13-07-2001 (dd-mm-yyyy)

	Copyright  2001 Vincent PRAT
*/

#include "../headers/NURBSSurface.h"

TrimmedNURBSSurface::~TrimmedNURBSSurface()
{
	// Free dynamically allocated memory
	delete pfUBasisCoefficients;
	delete pnUTessKnotSpan;
	delete pfDUBasisValues;
	delete pfUBasisValues;
	delete pfUKnots;

	delete pfVBasisCoefficients;
	delete pnVTessKnotSpan;
	delete pfDVBasisValues;
	delete pfVBasisValues;
	delete pfVKnots;

	delete pTextureCoordinates;
	delete pControlPoints;
	delete pVertexBuffer;
	delete pNormals;

	// Free each contour individually
	for ( int i=0; i<nNumberOfContours; i++ )
	{
		delete pContours[i];
	}
	delete pContours;
}

TrimmedNURBSSurface::TrimmedNURBSSurface( char *pcFileName )
{
	int i, j;
	
	// Create the I/O stream by opening file named pcFileName
	ifstream f( pcFileName );
	
	// Read in the file, see file format
	f >> anOrder[0] >> anOrder[1];
	f >> anNumberOfControlPoints[0] >> anNumberOfControlPoints[1];
	
	// Allocate memory for control points
	pControlPoints = new Vector4d[anNumberOfControlPoints[0]*anNumberOfControlPoints[1]];

	// Read control points
	for ( j=0; j<anNumberOfControlPoints[1]; j++)
	{
		for(i = 0; i < anNumberOfControlPoints[0]; i++)
		{
			f   >> pControlPoints[j*anNumberOfControlPoints[0]+i].x 
				>> pControlPoints[j*anNumberOfControlPoints[0]+i].y 
				>> pControlPoints[j*anNumberOfControlPoints[0]+i].z 
				>> pControlPoints[j*anNumberOfControlPoints[0]+i].w;
		}
	}
	
	// Allocate memory for knots
	pfUKnots = new float[ anOrder[0]+anNumberOfControlPoints[0] ];

	// Read U Knots
	for(i = 0; i < anOrder[0]+anNumberOfControlPoints[0]; i++)
	{
		f >> pfUKnots[i];
	}
	
	// Allocate memory for knots
	pfVKnots = new float[ anOrder[1]+anNumberOfControlPoints[1] ];

	// Read V Knots
	for(i = 0; i < anOrder[1]+anNumberOfControlPoints[1]; i++)
	{
		f >> pfVKnots[i];
	}
	
	f >> nNumberOfContours;
	
	pContours = NULL;

	// Any contour ?
	if (nNumberOfContours!=0)
	{
		// Allocate memory to store contour object pointers
		pContours = new LPCONTOUR2D[nNumberOfContours];
		
		// Read each contour
		for ( i=0; i<nNumberOfContours; i++)
		{
			pContours[i] = new Contour2D();
			
			f >> pContours[i]->nOrder;
			f >> pContours[i]->nNumberOfControlPoints;
			
			// Read control points
			pContours[i]->pControlPoints = new Vector4d[ pContours[i]->nNumberOfControlPoints ];
			for (j=0; j<pContours[i]->nNumberOfControlPoints; j++)
			{
				f   >> pContours[i]->pControlPoints[j].x
					>> pContours[i]->pControlPoints[j].y
					>> pContours[i]->pControlPoints[j].w;
			}
			
			// read knots
			pContours[i]->pfKnots = new float[ pContours[i]->nOrder + pContours[i]->nNumberOfControlPoints ];
			for (j=0; j<pContours[i]->nOrder + pContours[i]->nNumberOfControlPoints; j++)
			{
				f  >> pContours[i]->pfKnots[j];
			}
			
			// Prepare and tesselate contour
			pContours[i]->Initialise();
		}
	}

	// Prepare and tesselate surface
	Initialise();
}

void TrimmedNURBSSurface::Initialise()
{
	// Initialise the values not given at construction time
	nTesselation			= DEFAULT_TESSELATION_VALUE;

	anDegree[0]				= anOrder[0] - 1;
	anDegree[1]				= anOrder[1] - 1;

	anNumberOfKnots[0]		= anOrder[0] + anNumberOfControlPoints[0];
	anNumberOfKnots[1]		= anOrder[1] + anNumberOfControlPoints[1];

	anNumberOfKnotSpans[0]	= anNumberOfKnots[0] - 2*anDegree[0];
	anNumberOfKnotSpans[1]	= anNumberOfKnots[1] - 2*anDegree[1];

	// Allocate memory for arrays
	pfUBasisCoefficients	= NULL;
	pfVBasisCoefficients	= NULL;
	pfDUBasisValues			= NULL;
	pfDVBasisValues			= NULL;
	pnUTessKnotSpan			= NULL;
	pnVTessKnotSpan			= NULL;
	pfUBasisValues			= NULL;
	pfVBasisValues			= NULL;

	pTextureCoordinates		= NULL;
	pVertexBuffer			= NULL;
	pNormals				= NULL;

	pfUBasisCoefficients	= new float[ anOrder[0]*anOrder[0]*anNumberOfKnotSpans[0] ];
	pfVBasisCoefficients	= new float[ anOrder[1]*anOrder[1]*anNumberOfKnotSpans[1] ];
	pfDUBasisValues			= new float[ anOrder[0]*(nTesselation+1) ];
	pfDVBasisValues			= new float[ anOrder[1]*(nTesselation+1) ];	
	pnUTessKnotSpan			= new int[ nTesselation+1 ];	
	pnVTessKnotSpan			= new int[ nTesselation+1 ];
	pfUBasisValues			= new float[ anOrder[0]*(nTesselation+1) ];
	pfVBasisValues			= new float[ anOrder[1]*(nTesselation+1) ];
	
	pTextureCoordinates		= new Vector2d[ (nTesselation+1)*(nTesselation+1) ];
	pVertexBuffer			= new Vector4d[ (nTesselation+1)*(nTesselation+1) ];
	pNormals				= new Vector3d[ (nTesselation+1)*(nTesselation+1) ];

	// Pre-weight control points
	for ( int i=0; i<anNumberOfControlPoints[0]*anNumberOfControlPoints[1]; i++ )
	{
		pControlPoints[i].x *= pControlPoints[i].w;
		pControlPoints[i].y *= pControlPoints[i].w;
		pControlPoints[i].z *= pControlPoints[i].w;
	}

	// Evaluate the basis coefficients
	EvaluateBasisFunctionsCoefficients();

	// Then the basis function values at tesselated points
	EvaluateBasisFunctions();
	
	// Trim the surface if there are any contours
	bPerformTrimming = (nNumberOfContours != 0);

	// Pre-calculate surface points corresponding to the contours
	for ( i=0; i<nNumberOfContours; i++ )
	{
		pContours[i]->ComputeSurfacePoints( this );
	}

	// Tesselate the surface
	Tesselate();
}

void TrimmedNURBSSurface::SetTesselation( int newValue )
{
	// Check if the given value is different from the old one
	if ( newValue != nTesselation )
	{
		// Set tesselation, clip given value between 1 and maxTesselation
		nTesselation = ( newValue < 1 ? 1 : newValue );
		if ( nTesselation > DEFAULT_TESSELATION_VALUE )
		{
			nTesselation = DEFAULT_TESSELATION_VALUE;
		}

		// Free some arrays
		delete pfUBasisValues;
		delete pfVBasisValues;
		delete pfDUBasisValues;
		delete pfDVBasisValues;
		delete pnUTessKnotSpan;
		delete pnVTessKnotSpan;

		delete pTextureCoordinates;
		delete pVertexBuffer;
		delete pNormals;
		
		// Allocate the arrays with a new number of elements
		pfDUBasisValues		= new float[ anOrder[0]*(nTesselation+1) ];
		pfDVBasisValues		= new float[ anOrder[1]*(nTesselation+1) ];	
		pnUTessKnotSpan		= new int[ nTesselation+1 ];	
		pnVTessKnotSpan		= new int[ nTesselation+1 ];
		pfUBasisValues		= new float[ anOrder[0]*(nTesselation+1) ];
		pfVBasisValues		= new float[ anOrder[1]*(nTesselation+1) ];

		pVertexBuffer		= new Vector4d[ (nTesselation+1)*(nTesselation+1) ];
		pNormals			= new Vector3d[ (nTesselation+1)*(nTesselation+1) ];
		pTextureCoordinates = new Vector2d[ (nTesselation+1)*(nTesselation+1) ];

		// Recompute basis functions values
		EvaluateBasisFunctions();

		// Tesselate the surface
		Tesselate();
	}
}

void TrimmedNURBSSurface::RenderLOD( float xCam, float yCam, float zCam )
{
	int		i, j, newTesselationValue;
	long	offset;
	bool	stripInitialized;
	
	// Dynamic tesselation process

	// Get distance to view point
	float distance = sqrt(
		(xCam-vPosition.x)*(xCam-vPosition.x) +
		(yCam-vPosition.y)*(yCam-vPosition.y) +
		(zCam-vPosition.z)*(zCam-vPosition.z) );
	
	// Choose tesselation value
	if (distance<=CLOSE_DISTANCE)
		newTesselationValue = DEFAULT_TESSELATION_VALUE;
	else
		newTesselationValue = DEFAULT_TESSELATION_VALUE*FAR_DISTANCE/(distance-CLOSE_DISTANCE);
	
	if (newTesselationValue < 0)
		return;
	else if (newTesselationValue < 1)
		newTesselationValue = 1;
	else if (newTesselationValue > FAR_DISTANCE)
		newTesselationValue = FAR_DISTANCE;
	
	// And set it
	SetTesselation( newTesselationValue );

	offset = 0;
	stripInitialized = 0;

	if ( bPerformTrimming )
	{
		// Surface trimmed, take the holes into account
		j = 0; i = 0;
		
		glBegin( GL_TRIANGLE_STRIP );
		
		while ( j<nTesselation )
		{			
			// for each row
			if (i <= nTesselation)
			{
				// If we are on a valid column
				if (!stripInitialized)
				{
					// Strip not initialized or found a hole, end previous strip
					glEnd();

					// find the beginning of this strip
					while ( 
						( pVertexBuffer[offset].w >= 0 ) && 
						( i <= nTesselation ) )
					{
						offset++;
						i++;
					}

					// Begin the strip
					stripInitialized = 1;
					glBegin( GL_TRIANGLE_STRIP );
				}

				if (i<=nTesselation)
				{
					// if we have not reached the end of the column
					if ( ( pVertexBuffer[offset].w <= -1 ) &&
						 ( pVertexBuffer[offset+nTesselation+1].w <= -1 ) )

					{				
						// If the vertex is on surface, use it
						glNormal3f( 
							pNormals[offset].x,
							pNormals[offset].y,
							pNormals[offset].z );

						glVertex3f( 
							pVertexBuffer[offset].x,
							pVertexBuffer[offset].y,
							pVertexBuffer[offset].z );
						
						// Also use the vertex above it
						glNormal3f( 
							pNormals[offset+nTesselation+1].x,
							pNormals[offset+nTesselation+1].y,
							pNormals[offset+nTesselation+1].z );

						glVertex3f( 
							pVertexBuffer[offset+nTesselation+1].x,
							pVertexBuffer[offset+nTesselation+1].y,
							pVertexBuffer[offset+nTesselation+1].z );						
					}
					
					// Next vertex
					i++;
					offset++;
					
					if ( ( pVertexBuffer[offset].w >= 0 ) ||
						( pVertexBuffer[offset+nTesselation+1].w >= 0 ) )
					{
						// This is the end of a strip (next vertex is not on the surface)
						stripInitialized = 0;
					}
				}
			}
			else
			{
				// End of the column
				// Back to next row, first column
				i = 0;
				j++;
				stripInitialized = 0;
			}
		}
		glEnd();
	}
	else
	{
		// If the surface is not trimmed
		for (j=0; j<nTesselation; j++)
		{
			// Begin a strip
			glBegin( GL_TRIANGLE_STRIP );	

			// And push vertice until we reach the end of the line
			for (i=0; i<=nTesselation; i++)
			{
				glNormal3f( 
					pNormals[offset].x,
					pNormals[offset].y,
					pNormals[offset].z );
				
				glVertex3f( 
					pVertexBuffer[offset].x,
					pVertexBuffer[offset].y,
					pVertexBuffer[offset].z );
				
				glNormal3f( 
					pNormals[offset+nTesselation+1].x,
					pNormals[offset+nTesselation+1].y,
					pNormals[offset+nTesselation+1].z );
				
				glVertex3f( 
					pVertexBuffer[offset+nTesselation+1].x,
					pVertexBuffer[offset+nTesselation+1].y,
					pVertexBuffer[offset+nTesselation+1].z );
				
				// go to next column
				offset++;
			}

			// End this strip
			glEnd();
		}	
	}
}

void TrimmedNURBSSurface::Render()
{
	// Same as above, without dynamic tesselation

	int		i, j;
	long	offset;
	bool	stripInitialized;

	offset = 0;
	stripInitialized = 0;

	Tesselate();

	if ( bPerformTrimming )
	{
		// Surface trimmed, take the holes into account
		j = 0; i = 0;
		
		glBegin( GL_TRIANGLE_STRIP );
		
		while ( j<nTesselation )
		{			
			// for each row
			if (i <= nTesselation)
			{
				// If we are on a valid column
				if (!stripInitialized)
				{
					// Strip not initialized or found a hole, end previous strip
					glEnd();

					// find the beginning of this strip
					while ( 
						( pVertexBuffer[offset].w >= 0 ) && 
						( i <= nTesselation ) )
					{
						offset++;
						i++;
					}

					// Begin the strip
					stripInitialized = 1;
					glBegin( GL_TRIANGLE_STRIP );
				}

				if (i<=nTesselation)
				{
					// if we have not reached the end of the column
					if ( ( pVertexBuffer[offset].w <= -1 ) &&
						 ( pVertexBuffer[offset+nTesselation+1].w <= -1 ) )

					{				
						// If the vertex is on surface, use it
						glNormal3f( 
							pNormals[offset].x,
							pNormals[offset].y,
							pNormals[offset].z );

						glVertex3f( 
							pVertexBuffer[offset].x,
							pVertexBuffer[offset].y,
							pVertexBuffer[offset].z );
						
						// Also use the vertex above it
						glNormal3f( 
							pNormals[offset+nTesselation+1].x,
							pNormals[offset+nTesselation+1].y,
							pNormals[offset+nTesselation+1].z );

						glVertex3f( 
							pVertexBuffer[offset+nTesselation+1].x,
							pVertexBuffer[offset+nTesselation+1].y,
							pVertexBuffer[offset+nTesselation+1].z );						
					}
					
					// Next vertex
					i++;
					offset++;
					
					if ( ( pVertexBuffer[offset].w >= 0 ) ||
						( pVertexBuffer[offset+nTesselation+1].w >= 0 ) )
					{
						// This is the end of a strip (next vertex is not on the surface)
						stripInitialized = 0;
					}
				}
			}
			else
			{
				// End of the column
				// Back to next row, first column
				i = 0;
				j++;
				stripInitialized = 0;
			}
		}
		glEnd();
	}
	else
	{
		// If the surface is not trimmed
		for (j=0; j<nTesselation; j++)
		{
			// Begin a strip
			glBegin( GL_TRIANGLE_STRIP );	

			// And push vertice until we reach the end of the line
			for (i=0; i<=nTesselation; i++)
			{
				glNormal3f( 
					pNormals[offset].x,
					pNormals[offset].y,
					pNormals[offset].z );
				
				glVertex3f( 
					pVertexBuffer[offset].x,
					pVertexBuffer[offset].y,
					pVertexBuffer[offset].z );
				
				glNormal3f( 
					pNormals[offset+nTesselation+1].x,
					pNormals[offset+nTesselation+1].y,
					pNormals[offset+nTesselation+1].z );
				
				glVertex3f( 
					pVertexBuffer[offset+nTesselation+1].x,
					pVertexBuffer[offset+nTesselation+1].y,
					pVertexBuffer[offset+nTesselation+1].z );
				
				// go to next column
				offset++;
			}

			// End this strip
			glEnd();
		}	
	}
}

void TrimmedNURBSSurface::EvaluateBasisFunctionsCoefficients()
{
	int i, j, k;

	// For U direction

	// For each knot span
	for ( i=0; i<anNumberOfKnotSpans[0]; i++ )
	{
		// For m_nOrder polynomials
		for ( j=0; j<anOrder[0]; j++)
		{
			// For each coefficient of the polynomial
			for ( k=0; k<anOrder[0]; k++)
			{
				pfUBasisCoefficients[ ( i*anOrder[0] + j )*anOrder[0] + k ] = 
					ComputeCoefficient( pfUKnots, i+anDegree[0], i+j, anDegree[0], k );
			}
		}
	}

	// For V direction

	// For each knot span
	for ( i=0; i<anNumberOfKnotSpans[1]; i++ )
	{
		// For m_nOrder polynomials
		for ( j=0; j<anOrder[1]; j++)
		{
			// For each coefficient of the polynomial
			for ( k=0; k<anOrder[1]; k++)
			{
				pfVBasisCoefficients[ ( i*anOrder[1] + j )*anOrder[1] + k ] = 
					ComputeCoefficient( pfVKnots, i+anDegree[1], i+j, anDegree[1], k );
			}
		}
	}
}

float TrimmedNURBSSurface::ComputeCoefficient( float *pfKnots, int nInterval, int i, int p, int k )
{
	// Very simple recursive function, exactly based on the formula
	// giving the coefficients

	float fResult = 0.0f;

	if ( p==0 )
	{
		if ( i==nInterval )
		{
			fResult = 1.0f;
		}
	}
	else if ( k==0 )
	{
		if ( pfKnots[i+p]!=pfKnots[i] )
		{
			fResult -= pfKnots[i]*ComputeCoefficient( pfKnots, nInterval, i, p-1, 0 )/(pfKnots[i+p] - pfKnots[i]);
		}
		if ( pfKnots[i+p+1]!=pfKnots[i+1] )
		{
			fResult += pfKnots[i+p+1]*ComputeCoefficient( pfKnots, nInterval, i+1, p-1, 0 )/( pfKnots[i+p+1] - pfKnots[i+1] );
		}
	}
	else if ( k==p )
	{
		if ( pfKnots[i+p]!=pfKnots[i] )
		{
			fResult += ComputeCoefficient( pfKnots, nInterval, i, p-1, p-1 )/(pfKnots[i+p] - pfKnots[i]);
		}
		if ( pfKnots[i+p+1]!=pfKnots[i+1] )
		{
			fResult -= ComputeCoefficient( pfKnots, nInterval, i+1, p-1, p-1 )/( pfKnots[i+p+1] - pfKnots[i+1] );
		}
	}
	else if ( k>p )
	{
		fResult = 0.0f;
	}
	else
	{
		float C1, C2;

		if ( pfKnots[i+p]!=pfKnots[i] )
		{
			C1 = ComputeCoefficient( pfKnots, nInterval, i, p-1, k-1 );
			C2 = ComputeCoefficient( pfKnots, nInterval, i, p-1, k );

			fResult += ( C1 - pfKnots[i]*C2 )/( pfKnots[i+p]-pfKnots[i] );
		}
		if ( pfKnots[i+p+1]!=pfKnots[i+1] )
		{
			C1 = ComputeCoefficient( pfKnots, nInterval, i+1, p-1, k-1 );
			C2 = ComputeCoefficient( pfKnots, nInterval, i+1, p-1, k );

			fResult -= ( C1 - pfKnots[i+p+1]*C2 )/( pfKnots[i+p+1] - pfKnots[i+1] );
		}
	}

	return fResult;
}

void TrimmedNURBSSurface::EvaluateBasisFunctions()
{
	int i, j, k, idx;
	float u, uInc;
	
	// For U direction

	// Current knot span
	idx = 0;

	// Initial value of u
	u	= pfUKnots[anDegree[0]];

	// u increment
	uInc= ( pfUKnots[anNumberOfKnots[0]-anOrder[0]] - pfUKnots[anDegree[0]])/(nTesselation);
	
	// For each tesselated point
	for ( i=0; i<=nTesselation; i++ )
	{
		// Find the next valid knot span
		while ( (idx<anNumberOfKnots[0]-2*anDegree[0]-2) && (u>= pfUKnots[idx+anDegree[0]+1]) )
		{
			idx++;
		}
		
		// Memorise it
		pnUTessKnotSpan[i] = idx+anDegree[0];
		
		// Evaluate the basis function for this value of u
		// We use Horner's method to evaluated a polynomial
		for ( j=0; j<anOrder[0]; j++)
		{
			pfUBasisValues[ i*anOrder[0]+j ] = 
				pfUBasisCoefficients[ (idx*anOrder[0]+j)*anOrder[0] + anDegree[0] ];
			pfDUBasisValues[ i*anOrder[0]+j ] = 
				pfUBasisValues[ i*anOrder[0]+j ]*anDegree[0];
			
			for ( k=anDegree[0]-1; k>=0; k-- )
			{
				pfUBasisValues[ i*anOrder[0]+j ] = 
					pfUBasisValues[ i*anOrder[0]+j ] * u + 
					pfUBasisCoefficients[ (idx*anOrder[0]+j)*anOrder[0] + k ];
				if (k>0)
				{
					pfDUBasisValues[ i*anOrder[0]+j ] = 
						pfDUBasisValues[ i*anOrder[0]+j ] * u +
						pfUBasisCoefficients[ (idx*anOrder[0]+j)*anOrder[0] + k ] * k;
				}
			}
		}
		
		// Next paramter value
		u += uInc;
	}
	
	// For V direction (see comments above, same thing)

	float v, vInc;
	
	idx = 0;
	v	= pfVKnots[anDegree[1]];
	vInc= ( pfVKnots[anNumberOfKnots[1]-anOrder[1]] - pfVKnots[anDegree[1]])/(nTesselation);
	
	for ( i=0; i<=nTesselation; i++ )
	{
		while ( (idx<anNumberOfKnots[1]-2*anDegree[1]-2) && (v>=pfVKnots[idx+anDegree[1]+1]) )
		{
			idx++;
		}
		
		pnVTessKnotSpan[i] = idx+anDegree[1];
		
		for ( j=0; j<anOrder[1]; j++)
		{
			pfVBasisValues[ i*anOrder[1]+j ] = 
				pfVBasisCoefficients[ (idx*anOrder[1]+j)*anOrder[1] + anDegree[1] ];
			pfDVBasisValues[ i*anOrder[1]+j ] = 
				pfVBasisValues[ i*anOrder[1]+j ]*anDegree[1];
			
			for ( k=anDegree[1]-1; k>=0; k-- )
			{
				pfVBasisValues[ i*anOrder[1]+j ] = 
					pfVBasisValues[ i*anOrder[1]+j ] * v + 
					pfVBasisCoefficients[ (idx*anOrder[1]+j)*anOrder[1] + k ];
				
				if (k>0)
				{
					pfDVBasisValues[ i*anOrder[1]+j ] = 
						pfDVBasisValues[ i*anOrder[1]+j ] * v +
						pfVBasisCoefficients[ (idx*anOrder[1]+j)*anOrder[1] + k ] * k;
				}
			}
		}
		
		v += vInc;
	}
}

void TrimmedNURBSSurface::Tesselate()
{
	int m, n, k, l;
	float rhw;
	float u, v, uInc, vInc;
	float tu, tv;		
	float tuInc, tvInc;
	float fTemp;
	
	int anCurrentKnotSpan[2];
	int anBasisOffset[2], nControlPointOffset;
	float fBasisValue, fDBasisValue;
	
	long offset = 0;

	int actualContour = -1;
	
	Vector4d oTemp;
	Vector4d oTanU, oTanV;
	Vector3d oNormal;
	int nearestPoint, tempOffset, choosenOffset;
	
	Vector4d *pUTemp;
	Vector4d *pDUTemp;
	
	// Allocate memory for temp vectors
	pUTemp  = new Vector4d[ anOrder[0] ];
	pDUTemp = new Vector4d[ anOrder[0] ];
	
	// Texture coordinates
	tu = tv = 0;

	// Increments for texture coordinates
	tuInc = 1.0f / (float)(nTesselation);
	tvInc = 1.0f / (float)(nTesselation);

	// u and v starting values
	u   = pfUKnots[anDegree[0]];		
	v	= pfVKnots[anDegree[1]];

	// u and v increments
	uInc= ( pfUKnots[anNumberOfKnots[0]-anOrder[0]] - pfUKnots[anDegree[0]])/(nTesselation);
	vInc= ( pfVKnots[anNumberOfKnots[1]-anOrder[1]] - pfVKnots[anDegree[1]])/(nTesselation);
	
	// Offset of the point which is one column and one row before current point
	tempOffset = -1 - (nTesselation+1);

	// For each row of the mesh
	for ( n=0; n<=nTesselation; n++ )
	{
		// current knot span in U direction
		anCurrentKnotSpan[1] = pnVTessKnotSpan[n];
		anCurrentKnotSpan[0] = -1;

		// Reset V coordinates for texture
		tu = 0.0f;
		u  = pfUKnots[anDegree[0]];
			
		// offset in the precalculated basis functions array and control points array
		anBasisOffset[1] = n*anOrder[1];

		// For each point on the row m
		for ( m=0; m<=nTesselation; m++ )
		{			
			// Check if we are on the surface (for trimmed surfaces)
			if ( bPerformTrimming )
			{
				actualContour = -1;
				for (k=0; k<nNumberOfContours; k++)
				{
					if ( pContours[k]->IsInside( u, v ) )
					{
						// Remember the contour we are in
						actualContour = k;
					}
				}		
			}
			
			if ( anCurrentKnotSpan[0]!=pnUTessKnotSpan[m] )
			{
				// The knot span has changed on the n direction,
				anCurrentKnotSpan[0] = pnUTessKnotSpan[m];
						
				for ( k=0; k<=anDegree[0]; k++ )
				{
					nControlPointOffset = 
						(anCurrentKnotSpan[1] - anDegree[1])*anNumberOfControlPoints[0] + 
						(anCurrentKnotSpan[0] - anDegree[0]);
					
					// Clear temp values
					fBasisValue  = pfVBasisValues[anBasisOffset[1]];
					fDBasisValue = pfDVBasisValues[anBasisOffset[1]];

					// We assume that the x, y, z coordinates of the control points are pre-weighted !!
					pUTemp[k].x = fBasisValue*pControlPoints[ nControlPointOffset+k ].x;//*pControlPoints[ nControlPointOffset+k ].w;
					pUTemp[k].y = fBasisValue*pControlPoints[ nControlPointOffset+k ].y;//*pControlPoints[ nControlPointOffset+k ].w;
					pUTemp[k].z = fBasisValue*pControlPoints[ nControlPointOffset+k ].z;//*pControlPoints[ nControlPointOffset+k ].w;
					pUTemp[k].w = fBasisValue*pControlPoints[ nControlPointOffset+k ].w;
					
					pDUTemp[k].x = fDBasisValue*pControlPoints[ nControlPointOffset+k ].x;//*pControlPoints[ nControlPointOffset+k ].w;
					pDUTemp[k].y = fDBasisValue*pControlPoints[ nControlPointOffset+k ].y;//*pControlPoints[ nControlPointOffset+k ].w;
					pDUTemp[k].z = fDBasisValue*pControlPoints[ nControlPointOffset+k ].z;//*pControlPoints[ nControlPointOffset+k ].w;
					pDUTemp[k].w = fDBasisValue*pControlPoints[ nControlPointOffset+k ].w;						

					for ( l=1; l<=anDegree[1]; l++ )
					{
						nControlPointOffset += anNumberOfControlPoints[0];

						fBasisValue  = pfVBasisValues[anBasisOffset[1]+l];
						fDBasisValue = pfDVBasisValues[anBasisOffset[1]+l];

						pUTemp[k].x += fBasisValue*pControlPoints[ nControlPointOffset+k ].x;//*pControlPoints[ nControlPointOffset+k ].w;
						pUTemp[k].y += fBasisValue*pControlPoints[ nControlPointOffset+k ].y;//*pControlPoints[ nControlPointOffset+k ].w;
						pUTemp[k].z += fBasisValue*pControlPoints[ nControlPointOffset+k ].z;//*pControlPoints[ nControlPointOffset+k ].w;
						pUTemp[k].w += fBasisValue*pControlPoints[ nControlPointOffset+k ].w;

						pDUTemp[k].x += fDBasisValue*pControlPoints[ nControlPointOffset+k ].x;//*pControlPoints[ nControlPointOffset+k ].w;
						pDUTemp[k].y += fDBasisValue*pControlPoints[ nControlPointOffset+k ].y;//*pControlPoints[ nControlPointOffset+k ].w;
						pDUTemp[k].z += fDBasisValue*pControlPoints[ nControlPointOffset+k ].z;//*pControlPoints[ nControlPointOffset+k ].w;
						pDUTemp[k].w += fDBasisValue*pControlPoints[ nControlPointOffset+k ].w;						
					}
				}
			}
			
			// Offset in basis values array, V direction
			anBasisOffset[0] = m*anOrder[0];
			
			// Clear temp vectors
			fBasisValue = pfUBasisValues[ anBasisOffset[0] ];
			fDBasisValue = pfDUBasisValues[ anBasisOffset[0] ];
			
			oTemp.x = fBasisValue*pUTemp[0].x;
			oTemp.y = fBasisValue*pUTemp[0].y;
			oTemp.z = fBasisValue*pUTemp[0].z;
			oTemp.w = fBasisValue*pUTemp[0].w;
			
			oTanU.x = fBasisValue*pDUTemp[0].x;
			oTanU.y = fBasisValue*pDUTemp[0].y;
			oTanU.z = fBasisValue*pDUTemp[0].z;
			oTanU.w = fBasisValue*pDUTemp[0].w;
			
			oTanV.x = fDBasisValue*pUTemp[0].x;
			oTanV.y = fDBasisValue*pUTemp[0].y;
			oTanV.z = fDBasisValue*pUTemp[0].z;
			oTanV.w = fDBasisValue*pUTemp[0].w;
			
			// Sum the U basis functions and determine coordinates, tangents
			for ( k=1; k<=anDegree[0]; k++ )
			{
				fBasisValue = pfUBasisValues[ anBasisOffset[0] + k ];
				fDBasisValue = pfDUBasisValues[ anBasisOffset[0] + k ];
				
				oTemp.x += fBasisValue*pUTemp[k].x;
				oTemp.y += fBasisValue*pUTemp[k].y;
				oTemp.z += fBasisValue*pUTemp[k].z;
				oTemp.w += fBasisValue*pUTemp[k].w;
				
				oTanU.x += fBasisValue*pDUTemp[k].x;
				oTanU.y += fBasisValue*pDUTemp[k].y;
				oTanU.z += fBasisValue*pDUTemp[k].z;
				oTanU.w += fBasisValue*pDUTemp[k].w;
				
				oTanV.x += fDBasisValue*pUTemp[k].x;
				oTanV.y += fDBasisValue*pUTemp[k].y;
				oTanV.z += fDBasisValue*pUTemp[k].z;
				oTanV.w += fDBasisValue*pUTemp[k].w;
			}
			
			// value to bring oTemp from 4D to 3D
			rhw = 1.0f/oTemp.w;
			
			oTemp.x *= rhw; 
			oTemp.y *= rhw;
			oTemp.z *= rhw;

			// Bring tangents from 4D to 3D
			oTanU.x = (oTanU.x - oTanU.w*oTemp.x) * rhw;
			oTanU.y = (oTanU.y - oTanU.w*oTemp.y) * rhw;
			oTanU.z = (oTanU.z - oTanU.w*oTemp.z) * rhw;
			
			oTanV.x = (oTanV.x - oTanV.w*oTemp.x) * rhw;
			oTanV.y = (oTanV.y - oTanV.w*oTemp.y) * rhw;
			oTanV.z = (oTanV.z - oTanV.w*oTemp.z) * rhw;
			
			// Determine normal (cross product of tanU and tanV)
			oNormal.x = (oTanU.y*oTanV.z - oTanV.y*oTanU.z);
			oNormal.y = (oTanU.z*oTanV.x - oTanV.z*oTanU.x);
			oNormal.z = (oTanV.y*oTanU.x - oTanU.y*oTanV.x);
			
			// Normalize its lenght
			fTemp = sqrt( oNormal.x*oNormal.x + oNormal.y*oNormal.y + oNormal.z*oNormal.z );
			oNormal.x /= fTemp;
			oNormal.y /= fTemp;
			oNormal.z /= fTemp;

			// save current vertex in the vertex buffer
			pVertexBuffer[offset].x = oTemp.x;
			pVertexBuffer[offset].y = oTemp.y;
			pVertexBuffer[offset].z = oTemp.z;
			// Save the contour this vertex is in.
			// -1 if the vertex is on the surface
			pVertexBuffer[offset].w = actualContour;
			
			// Save normal
			pNormals[offset].x = oNormal.x;
			pNormals[offset].y = oNormal.y;
			pNormals[offset].z = oNormal.z;
			
			// texture coordinates
			pTextureCoordinates[offset].x = tu;
			pTextureCoordinates[offset].y = tv;			
			
			// Refine the coordinates of the point if possible
			if ( (bPerformTrimming) && (m>1) && (n>1) && (pVertexBuffer[tempOffset].w == -1) )
			{
				// previous vertex is on the curve.
				
				// Is it on the border of a hole ?
				if ( 
					( pVertexBuffer[tempOffset-1].w >= 0 ) ||
					( pVertexBuffer[tempOffset+1].w >= 0 ) ||
					( pVertexBuffer[tempOffset-(nTesselation+1)-1].w >= 0 ) ||
					( pVertexBuffer[tempOffset-(nTesselation+1)+1].w >= 0 ) ||
					( pVertexBuffer[tempOffset-(nTesselation+1)  ].w >= 0 ) ||
					( pVertexBuffer[tempOffset+(nTesselation+1)-1].w >= 0 ) ||
					( pVertexBuffer[tempOffset+(nTesselation+1)+1].w >= 0 ) ||
					( pVertexBuffer[tempOffset+(nTesselation+1)  ].w >= 0 ) )
				{
					// If so, refine its coordinates
					
					if ( pVertexBuffer[tempOffset-1].w >= 0 )
						choosenOffset = tempOffset-1;
					else if	( pVertexBuffer[tempOffset+1].w >= 0 )
						choosenOffset = tempOffset+1;
					else if	( pVertexBuffer[tempOffset-(nTesselation+1)-1].w >= 0 )
						choosenOffset = tempOffset-(nTesselation+1)-1;
					else if	( pVertexBuffer[ tempOffset-(nTesselation+1)+1].w >= 0 )
						choosenOffset = tempOffset-(nTesselation+1)+1;
					else if	( pVertexBuffer[ tempOffset-(nTesselation+1) ].w >= 0 )
						choosenOffset = tempOffset-(nTesselation+1);
					else if	( pVertexBuffer[ tempOffset+(nTesselation+1)-1].w >= 0 )
						choosenOffset = tempOffset+(nTesselation+1)-1;
					else if	( pVertexBuffer[ tempOffset+(nTesselation+1)+1].w >= 0 )
						choosenOffset = tempOffset+(nTesselation+1)+1;
					else if	( pVertexBuffer[ tempOffset+(nTesselation+1)  ].w >= 0 )
						choosenOffset = tempOffset+(nTesselation+1);
					
					// Save coordinates
					nearestPoint = pContours[(int)pVertexBuffer[ choosenOffset ].w]->FindNearestPoint( 
						u, v,
						pfUKnots[0], pfUKnots[anNumberOfKnots[0]-1],
						pfVKnots[0], pfVKnots[anNumberOfKnots[1]-1] );
					pVertexBuffer[tempOffset].x = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].x;
					pVertexBuffer[tempOffset].y = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].y;
					pVertexBuffer[tempOffset].z = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].z;
				}
			}

			// Next vertex
			tu += tuInc;
			u  += uInc;
			offset++;
			tempOffset++;
		}
		tv += tvInc;
		v  += vInc;
	}

	delete pUTemp;
	delete pDUTemp;

	if ( (bPerformTrimming) && (nTesselation > 3) )
	{
		// Smooth borders of the contours for trimmed surfaces
		// Only a few rows and columns are remaining for this pass

		// First row
		v = pfVKnots[anDegree[1]];
		u = pfUKnots[anDegree[0]] + uInc;
		offset = 1;

		for (m=1; m<nTesselation; m++)
		{
			if (pVertexBuffer[offset].w == -1)
			{
				// This vertex is on the curve.					
				
				// Is it on the border of a hole ?
				if ( 
					( pVertexBuffer[offset-1].w >= 0 ) ||
					( pVertexBuffer[offset+1].w >= 0 ) )
				{
					// If so, refine its coordinates
					
					if ( pVertexBuffer[offset-1].w >= 0 )
						choosenOffset = offset-1;
					else if	( pVertexBuffer[offset+1].w >= 0 )
						choosenOffset = offset+1;
					
					// Save coordinates
					nearestPoint = pContours[(int)pVertexBuffer[ choosenOffset ].w]->FindNearestPoint( 
						u, v,
						pfUKnots[0], pfUKnots[anNumberOfKnots[0]-1],
						pfVKnots[0], pfVKnots[anNumberOfKnots[1]-1] );
					pVertexBuffer[offset].x = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].x;
					pVertexBuffer[offset].y = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].y;
					pVertexBuffer[offset].z = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].z;
				}
			}
			offset++;
			u+=uInc;
		}

		// Last row
		v = pfVKnots[anDegree[1]] + (nTesselation)*vInc;
		u = pfUKnots[anDegree[0]] + uInc;
		offset = (nTesselation)*(nTesselation+1) + 1;

		for ( m=1; m<nTesselation; m++ )
		{
			if (pVertexBuffer[offset].w == -1)
			{
				// This vertex is on the curve.					
				
				// Is it on the border of a hole ?
				if ( 
					( pVertexBuffer[offset-1].w >= 0 ) ||
					( pVertexBuffer[offset+1].w >= 0 ) )
				{
					// If so, refine its coordinates
					
					if ( pVertexBuffer[offset-1].w >= 0 )
						choosenOffset = offset-1;
					else if	( pVertexBuffer[offset+1].w >= 0 )
						choosenOffset = offset+1;
					else if	( pVertexBuffer[offset-(nTesselation+1)-1].w >= 0 )
						choosenOffset = offset-(nTesselation+1)-1;
					else if	( pVertexBuffer[ offset-(nTesselation+1)+1].w >= 0 )
						choosenOffset = offset-(nTesselation+1)+1;
					else if	( pVertexBuffer[ offset-(nTesselation+1) ].w >= 0 )
						choosenOffset = offset-(nTesselation+1);
					
					// Save coordinates
					nearestPoint = pContours[(int)pVertexBuffer[ choosenOffset ].w]->FindNearestPoint( 
						u, v,
						pfUKnots[0], pfUKnots[anNumberOfKnots[0]-1],
						pfVKnots[0], pfVKnots[anNumberOfKnots[1]-1] );
					pVertexBuffer[offset].x = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].x;
					pVertexBuffer[offset].y = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].y;
					pVertexBuffer[offset].z = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].z;
				}
			}
			u  += uInc;
			offset++;
		}

		// First column
		v = pfVKnots[anDegree[1]] + vInc;
		u = pfUKnots[anDegree[0]];
		offset = (nTesselation+1);

		for ( n=1; n<nTesselation; n++ )
		{
			if (pVertexBuffer[offset].w == -1)
			{
				// This vertex is on the curve.					
				
				// Is it on the border of a hole ?
				if ( 
					( pVertexBuffer[offset-(nTesselation+1)  ].w >= 0 ) ||
					( pVertexBuffer[offset+(nTesselation+1)  ].w >= 0 ) )
				{
					// If so, refine its coordinates
					
					if	( pVertexBuffer[ offset-(nTesselation+1) ].w >= 0 )
						choosenOffset = offset-(nTesselation+1);
					else if	( pVertexBuffer[ offset+(nTesselation+1)  ].w >= 0 )
						choosenOffset = offset+(nTesselation+1);
					
					// Save coordinates
					nearestPoint = pContours[(int)pVertexBuffer[ choosenOffset ].w]->FindNearestPoint( 
						u, v,
						pfUKnots[0], pfUKnots[anNumberOfKnots[0]-1],
						pfVKnots[0], pfVKnots[anNumberOfKnots[1]-1] );
					pVertexBuffer[offset].x = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].x;
					pVertexBuffer[offset].y = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].y;
					pVertexBuffer[offset].z = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].z;
				}
			}
			v  += vInc;
			offset += nTesselation+1;
		}

		// Last column
		v = pfVKnots[anDegree[1]] + vInc;
		u = pfUKnots[anDegree[0]] + nTesselation*uInc;
		offset = (nTesselation+1) + nTesselation;

		for ( n=1; n<nTesselation; n++ )
		{
			if (pVertexBuffer[offset].w == -1)
			{
				// This vertex is on the curve.					
				
				// Is it on the border of a hole ?
				if ( 
					( pVertexBuffer[offset-(nTesselation+1)  ].w >= 0 ) ||
					( pVertexBuffer[offset+(nTesselation+1)  ].w >= 0 ) )
				{
					// If so, refine its coordinates
					if	( pVertexBuffer[ offset-(nTesselation+1) ].w >= 0 )
						choosenOffset = offset-(nTesselation+1);
					else if	( pVertexBuffer[ offset+(nTesselation+1)  ].w >= 0 )
						choosenOffset = offset+(nTesselation+1);
					
					// Save coordinates
					nearestPoint = pContours[(int)pVertexBuffer[ choosenOffset ].w]->FindNearestPoint( 
						u, v,
						pfUKnots[0], pfUKnots[anNumberOfKnots[0]-1],
						pfVKnots[0], pfVKnots[anNumberOfKnots[1]-1] );
					pVertexBuffer[offset].x = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].x;
					pVertexBuffer[offset].y = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].y;
					pVertexBuffer[offset].z = pContours[(int)pVertexBuffer[ choosenOffset ].w]->pSurfacePoints[nearestPoint].z;
				}
			}
			v  += vInc;
			offset += nTesselation+1;
		}
	}	
}

float TrimmedNURBSSurface::EvaluateBasisFunction( int i, int k, float u, float *pfKnots )
{
	float fResult = 0.0;
	
	if ( k==0 )
	{
		if ( (u < pfKnots[i+1]) && (u >= pfKnots[i]) )
		{
			fResult = 1.0;
		}
	}
	else
	{
		float D1 = pfKnots[i+k] - pfKnots[i];
		float D2 = pfKnots[i+k+1] - pfKnots[i+1];
		
		if ( D1!=0 )
		{
			float C1 = EvaluateBasisFunction( i, k-1, u, pfKnots );
			fResult += C1*(float)(u-pfKnots[i])/D1;
		}
		if ( D2!=0 )
		{
			float C2 = EvaluateBasisFunction( i+1, k-1, u, pfKnots );
			fResult += C2*(float)(pfKnots[i+k+1]-u)/D2;
		}
	}
	
	return fResult;
}